@@ -1,27 +1,31 @@
-/******************************************************
+/* 
+ * limiter~: limit/compress signals
  *
- * zexy - implementation file
+ * (c) 1999-2011 IOhannes m zmÃ¶lnig, forum::fÃŒr::umlÃ€ute, institute of electronic music and acoustics (iem)
  *
- * copyleft (c) IOhannes m zmlnig
- *
- *   1999:forum::fr::umlute:2004
- *
- *   institute of electronic music and acoustics (iem)
- *
- ******************************************************
- *
- * license: GNU General Public License v.2
- *
- ******************************************************/
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License
+ * as published by the Free Software Foundation; either version 2
+ * of the License, or (at your option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License along
+ * with this program.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
 
 
 /*
-	--------------------------------- limiter/compressor ---------------------------------
+  --------------------------------- limiter/compressor ---------------------------------
 	
-	for details on how it works watch out for   "http://iem.kug.ac.at/~zmoelnig/pd"
-	...and search for "limiter"
+  for details on how it works watch out for   "http://iem.kug.ac.at/~zmoelnig/pd"
+  ...and search for "limiter"
 
-	mail2me4more!n4m8ion : zmoelnig@iem.kug.ac.at
+  mail2me4more!n4m8ion : zmoelnig@iem.kug.ac.at
 */
 
 /* 
@@ -29,7 +33,7 @@
    the limiter is based on Falkner's thesis 
    "Entwicklung eines digitalen Stereo-limiters mit Hilfe des Signalprozessors DSP56001" pp.14
 
-   2108:forum::fr::umlute:1999		all rights reserved and no warranties...
+   2108:forum::fÃŒr::umlÃ€ute:1999		all rights reserved and no warranties...
 
    see GNU-license for details
 */
@@ -40,44 +44,73 @@
 
 #include "zexy.h"
 
+/* log2 */
 #define LN2 .69314718056
-#define SINC1 .822462987
-#define SINC2 .404460777
-#define SINC3 -.188874003
-#define SINC4 -.143239449
-#define SINC5 .087796546
-#define SINC6 .06917082
-#define SINC7 -.041349667
-#define SINC8 -.030578954
-#define SINC9 .013226276
+
+
+/* hmm, where do these values come from exactly?
+   we are doing a 3* oversampling, so we need to calculate the samples that are +-1/3 off the current sample
+   sinc([13 10 7 4 1 -2 -5 -8 -11]/3) * window
+   window=cosine window
+
+*/
+
+/*
+#define SINC[4] .822462987
+#define SINC[3] .404460777
+#define SINC[5] -.188874003
+#define SINC[2] -.143239449
+#define SINC[6] .087796546
+#define SINC[1] .06917082
+#define SINC[7] -.041349667
+#define SINC[0] -.030578954
+#define SINC[8] .013226276
+*/
 
 #define BUFSIZE 128
 #define XTRASAMPS 9
 #define TABLESIZE 512 /* compressor table */
 
+static t_sample SINC[9];
+
+#define PI  3.1415926535897932384626433832795029L  /* pi */
+
+static void init_sinc(void) {
+  /*  calculate the sinc (windowed with a cosine) */
+  int i=0;
+  for(i=0; i<9; i++) {
+    long double t=(3.*i - 11.)/3.;
+    long double v=cos(t*PI/10.)*sin(PI*t)/(PI*t);
+    SINC[i]=v;
+  }
+}
+
 /* ------------------------------------------------------------------------------------ */
-// first define the structs...
+/* first define the structs... */
 
 static t_class *limiter_class;
 
 typedef struct _limctl
-{		// variables changed by user
-  float limit, hold_samples, change_of_amplification;
+{		
+  /* variables changed by user */
+  t_float limit;
+  t_float hold_samples;
+  t_float change_of_amplification;
 } t_limctl;
 
 typedef struct _cmpctl
 {
-  float treshold, ratio;		// uclimit is the very same is the limiter1-limit (decalculated relative to our treshold)
-  float uclimit, climit_inverse;	// climit == compressed limit (uclimit == uncompressed limit)
+  t_float treshold, ratio;		/* uclimit is the very same is the limiter1-limit (decalculated relative to our treshold) */
+  t_float uclimit, climit_inverse;	/* climit == compressed limit (uclimit == uncompressed limit) */
 
-  float limiter_limit;		// start limiting (stop compressing); == tresh/limit;
+  t_float limiter_limit;		/* start limiting (stop compressing); == tresh/limit; */
 
-  float treshdB, oneminusratio;
+  t_float treshdB, oneminusratio;
 } t_cmpctl;
 
 typedef struct _inbuf
 {
-  float*		ringbuf;
+  t_sample*	ringbuf;
   int		buf_position;
 } t_inbuf;
 
@@ -87,19 +120,19 @@ typedef struct _limiter
 
   int		number_of_inlets, s_n;
 
-  // variables changed by process
-
-  float		amplification;
-  float		samples_left, still_left;
+  /* variables changed by process */
+  t_sample	amplification;
+  t_float	samples_left, still_left;
 
   int		mode;
 
   t_limctl	*val1, *val2;
   t_cmpctl	*cmp;
 
-  // note :	limit is not the same for val1 & val2 :
-  //			at val1 it is the limit of the INPUT_VALUE
-  //			at val2 it is the limit for the AMPLIFICATION (in fact it is abs_limit1/abs_limit2)
+  /* note :	limit is not the same for val1 & val2 :
+   *			at val1 it is the limit of the INPUT_VALUE
+   *			at val2 it is the limit for the AMPLIFICATION (in fact it is abs_limit1/abs_limit2)
+   */
 
   t_inbuf*	in;
   int		buf_size;
@@ -107,14 +140,15 @@ typedef struct _limiter
 } t_limiter;
 
 /* ------------------------------------------------------------------------------------ */
-// then do the message - thing
+/* then do the message - thing */
 
-// do the user settings
+/* do the user settings */
 
-// calcs
+/* calcs */
 static t_float calc_holdsamples(t_float htime, int buf)
-{	// hold_time must be greater than buffer_time to make sure that any peak_sample is amplified with its own factor
-  float min_hold = buf / sys_getsr();
+{	
+  /* hold_time must be greater than buffer_time to make sure that any peak_sample is amplified with its own factor */
+  t_float min_hold = buf / sys_getsr();
   return (0.001 * sys_getsr() * ((htime > min_hold)?htime:((min_hold > 50)?min_hold:50)));
 }
 
@@ -135,12 +169,12 @@ static void set_uclimit(t_limiter *x)
   c->oneminusratio = 1. - ratio;
 }
 
-// settings
+/* settings */
 
-static void set_treshold(t_limiter *x, float treshold)
+static void set_treshold(t_limiter *x, t_float treshold)
 {
   t_cmpctl *c = x->cmp;
-  float tresh = dbtorms (treshold);
+  t_float tresh = dbtorms (treshold);
   if (tresh > x->val1->limit) tresh = x->val1->limit;
 
   c->treshold = tresh;
@@ -148,7 +182,7 @@ static void set_treshold(t_limiter *x, float treshold)
   set_uclimit(x);
 }
 
-static void set_ratio(t_limiter *x, float ratio)
+static void set_ratio(t_limiter *x, t_float ratio)
 {
   if (ratio < 0) ratio = 1;
   x->cmp->ratio = ratio;
@@ -156,7 +190,7 @@ static void set_ratio(t_limiter *x, float ratio)
   set_uclimit(x);
 }
 
-static void set_mode(t_limiter *x, float  mode)
+static void set_mode(t_limiter *x, t_float mode)
 {
   int modus = mode;
 
@@ -174,7 +208,6 @@ static void set_mode(t_limiter *x, float  mode)
     x->mode = LIMIT0;
     break;
   }
-  //  post("mode set to %d", x->mode);
 }
 
 static void set_LIMIT(t_limiter *x)
@@ -192,8 +225,9 @@ static void set_COMPRESS(t_limiter *x)
   set_mode(x, COMPRESS);
 }
 
-static void set_bufsize(t_limiter *x, float size)
-{ // this is really unneeded...and for historical reasons only
+static void set_bufsize(t_limiter *x, int size)
+{ 
+  /* this is really unneeded...and for historical reasons only */
   if (size < BUFSIZE) size = BUFSIZE;
   x->buf_size = size + XTRASAMPS;
 }
@@ -218,8 +252,8 @@ static void set_limits(t_limiter *x, t_floatarg limit1, t_floatarg limit2)
 
   if (lim2 < lim1)
     {
-      lim2 = 2*lim1;	// this is to prevent lim2 (which should trigger the FAST regulation)
-      x->mode = 0;	// to underrun the SLOW regulation; this would cause distortion
+      lim2 = 2*lim1;	/* this is to prevent lim2 (which should trigger the FAST regulation) */
+      x->mode = 0;	/* to underrun the SLOW regulation; this would cause distortion */
     }
 
   x->val1->limit = lim1;
@@ -276,7 +310,7 @@ static void reset(t_limiter *x)
   x->amplification = 1.;
 }
 
-// verbose
+/* verbose */
 static void status(t_limiter *x)
 {
   t_limctl *v1 = x->val1;
@@ -315,7 +349,7 @@ static void status(t_limiter *x)
   }
 }
 
-static void helper(t_limiter *x)
+static void limiter_tilde_helper(t_limiter *x)
 {
   post("\n\n%c %d-channel limiter-object: mode %d", HEARTSYMBOL, x->number_of_inlets, x->mode);
   poststring("\n'mode <mode>'\t\t\t: (0_limiter, 1_crack-limiter, 2_compressor)");
@@ -334,7 +368,7 @@ static void helper(t_limiter *x)
 	       "\n'set2 <limit2><htime2><rtime2>'\t: set crack-limiter");
     break;
   case COMPRESS:
-    poststring("\n'ratio <compressratio>'\t\t: set compressratio (0.5 instead of 1:2)"
+    poststring("\n'ratio <compressratio>'\t\t: set compressratio (Åœ0.5Åœ instead of Åœ1:2Åœ)"
 	       "\n'treshold <treshold>'\t\t: set treshold of the compressor"
 	       "\n'compress <limit><treshold><ratio>'\t: set compressor"
 	       "\n..........note that <limit> is the same for COMPRESSOR and LIMITER..........");
@@ -351,68 +385,95 @@ static void helper(t_limiter *x)
 
 
 /* ------------------------------------------------------------------------------------ */
-// now do the dsp - thing							        //
+/* now do the dsp - thing							                                                  */
 /* ------------------------------------------------------------------------------------ */
 
 static t_int *oversampling_maxima(t_int *w)
 {
   t_limiter *x = (t_limiter *)w[1];
   t_inbuf *buf = (t_inbuf *)w[2];
-  t_float *in	 = (t_float *)w[3];
-  t_float *out = (t_float *)w[4];
+  t_sample *in	 = (t_sample *)w[3];
+  t_sample *out = (t_sample *)w[4];
 
   int n = x->s_n;
   int bufsize = x->buf_size;
 
   int i = buf->buf_position;
 
-  t_float *vp = buf->ringbuf, *ep = vp + bufsize, *bp = vp + XTRASAMPS + i;
+  t_sample *vp = buf->ringbuf, *ep = vp + bufsize, *bp = vp + XTRASAMPS + i;
 
   i += n;
 
   while (n--)
     {
-      t_float os1, os2, max;
-      t_float last4, last3, last2, last1, sinccurrent, current, next1, next2, next3, next4;
+      t_sample os1, os2, max;
+      t_sample last4, last3, last2, last1, sinccurrent, current, next1, next2, next3, next4;
 
       if (bp == ep)
-	{
-	  vp[0] = bp[-9];
-	  vp[1] = bp[-8];
-	  vp[2] = bp[-7];
-	  vp[3] = bp[-6];
-	  vp[4] = bp[-5];
-	  vp[5] = bp[-4];
-	  vp[6] = bp[-3];
-	  vp[7] = bp[-2];
-	  vp[8] = bp[-1];
+        {
+          vp[0] = bp[-9];
+          vp[1] = bp[-8];
+          vp[2] = bp[-7];
+          vp[3] = bp[-6];
+          vp[4] = bp[-5];
+          vp[5] = bp[-4];
+          vp[6] = bp[-3];
+          vp[7] = bp[-2];
+          vp[8] = bp[-1];
+          
+          bp = vp + XTRASAMPS;
+          i -= bufsize - XTRASAMPS;
+        }
 
-	  bp = vp + XTRASAMPS;
-	  i -= bufsize - XTRASAMPS;
-	}
 
-      os1= fabsf(SINC8 * (last4 = bp[-8]) +
-		 SINC6 * (last3 = bp[-7]) +
-		 SINC4 * (last2 = bp[-6]) +
-		 SINC2 * (last1 = bp[-5]) +
-		 (sinccurrent = SINC1 * (current = bp[-4])) +
-		 SINC3 * (next1 = bp[-3]) +
-		 SINC5 * (next2 = bp[-2]) +
-		 SINC7 * (next3 = bp[-1]) +
-		 SINC9 * (next4 = bp[0]));
+      last4 = bp[-8];
+      last3 = bp[-7];
+      last2 = bp[-6];
+      last1 = bp[-5];
+      current = bp[-4];
+      next1 = bp[-3];
+      next2 = bp[-2];
+      next3 = bp[-1];
+      next4 = bp[0];
 
-      os2= fabsf(SINC8 * next4 +
-		 SINC4 * next3 +
-		 SINC6 * next2 +
-		 SINC2 * next1 +
-		 sinccurrent +
-		 SINC3 * last1 +
-		 SINC5 * last2 +
-		 SINC7 * last3 +
-		 SINC9 * last4);
+      sinccurrent = SINC[4] * current;
+
+      os1= fabsf(SINC[0] * last4 +
+                 SINC[1] * last3 +
+                 SINC[2] * last2 +
+                 SINC[3] * last1 +
+                 sinccurrent +
+                 SINC[5] * next1 +
+                 SINC[6] * next2 +
+                 SINC[7] * next3 +
+                 SINC[8] * next4);
+
+      os2= fabsf(SINC[0] * next4 +
+                 SINC[1] * next3 +
+                 SINC[2] * next2 +
+                 SINC[3] * next1 +
+                 sinccurrent +
+                 SINC[5] * last1 +
+                 SINC[6] * last2 +
+                 SINC[7] * last3 +
+                 SINC[8] * last4);
 
       max = fabsf(current);
 
+#if 0
+      if(max>1. || os1>1. || os2>1.)
+        post("%f %f %f\t%f %f %f %f %f %f %f %f %f", max, os1, os2,
+             last4,
+             last3,
+             last2,
+             last1,
+             current,
+             next1,
+             next2,
+             next3,
+             next4
+             );
+#endif
       if (max < os1) 
 	{
 	  max = os1;
@@ -436,16 +497,16 @@ static t_int *limiter_perform(t_int *w)
   t_limiter *x=(t_limiter *)w[1];
   int n = x->s_n;
 
-  t_float *in	= (t_float *)w[2];
-  t_float *out= (t_float *)w[3];
+  t_sample *in	= (t_sample *)w[2];
+  t_sample *out= (t_sample *)w[3];
 
   t_limctl *v1	= (t_limctl *)(x->val1);
   t_limctl *v2	= (t_limctl *)(x->val2);
   t_cmpctl *c     = (t_cmpctl *)(x->cmp);
 
-  // now let's make things a little bit faster
+  /* now let's make things a little bit faster */
 
-  // these must not be changed by process
+  /* these MUST NOT be changed by process */
   const t_float limit		= v1->limit;
   const t_float holdlong	= v1->hold_samples;
   const t_float coa_long	= v1->change_of_amplification;
@@ -460,12 +521,12 @@ static t_int *limiter_perform(t_int *w)
 
   t_float oneminusratio = c->oneminusratio;
 
-  // these will be changed by process
+  /* these will be changed by process */
   t_float amp				= x->amplification;
   t_float samplesleft		= x->samples_left;
   t_float stillleft		= x->still_left;
 
-  // an intern variable...
+  /* an intern variable... */
   t_float max_val;
 	
   switch (x->mode) {
@@ -474,22 +535,22 @@ static t_int *limiter_perform(t_int *w)
       {
 	max_val = *in;
 
-	// the MAIN routine for the 1-treshold-limiter
+	/* the MAIN routine for the 1-treshold-limiter */
 
 	if ((max_val * amp) > limit)
 	  {
 	    amp = limit / max_val;
 	    samplesleft = holdlong;
 	  } else
-	    {
-	      if (samplesleft > 0)
-		{
-		  samplesleft--;
-		} else
-		  {
-		    if ((amp *= coa_long) > 1) amp = 1;
-		  }
-	    }
+          {
+            if (samplesleft > 0)
+              {
+                samplesleft--;
+              } else
+              {
+                if ((amp *= coa_long) > 1) amp = 1;
+              }
+          }
 
 	*out++ = amp;
 	*in++ = 0;
@@ -499,35 +560,35 @@ static t_int *limiter_perform(t_int *w)
     while (n--)
       {
 	max_val = *in;
-	// the main routine 2
+	/* the main routine 2 */
 
 	if ((max_val * amp) > limit)
 	  {
 	    samplesleft = ((amp = (limit / max_val)) < alimit)?holdshort:holdlong;
 	    stillleft = holdlong;
 	  } else
-	    {
-	      if (samplesleft > 0)
-		{
-		  samplesleft--;
-		  stillleft--;
-		} else
-		  {
-		    if (amp < alimit)
-		      {
-			if ((amp *= coa_short) > 1) amp = 1;
-		      } else
-			{
-			  if (stillleft > 0)
-			    {
-			      samplesleft = stillleft;
-			    } else
-			      {
-				if ((amp *= coa_long) > 1) amp = 1;
-			      }
-			}
-		  }
-	    }
+          {
+            if (samplesleft > 0)
+              {
+                samplesleft--;
+                stillleft--;
+              } else
+              {
+                if (amp < alimit)
+                  {
+                    if ((amp *= coa_short) > 1) amp = 1;
+                  } else
+                  {
+                    if (stillleft > 0)
+                      {
+                        samplesleft = stillleft;
+                      } else
+                      {
+                        if ((amp *= coa_long) > 1) amp = 1;
+                      }
+                  }
+              }
+          }
 	*out++ = amp;
 	*in++ = 0;
       }
@@ -538,7 +599,7 @@ static t_int *limiter_perform(t_int *w)
       {
 	max_val = *in;
 
-	// the MAIN routine for the compressor (very similar to the 1-treshold-limiter)
+	/* the MAIN routine for the compressor (very similar to the 1-treshold-limiter) */
 
 	if (max_val * amp > tresh) {
 	  amp = tresh / max_val;
@@ -548,9 +609,9 @@ static t_int *limiter_perform(t_int *w)
 	  else if ((amp *= coa_long) > 1) amp = 1;
 
 	if (amp < 1.)
-	  if (amp > uclimit)  // amp is still UnCompressed  uclimit==limitIN/tresh;
+	  if (amp > uclimit)  /* amp is still UnCompressed  uclimit==limitIN/tresh; */
 	    *out++ = pow(amp, oneminusratio);
-	  else *out++ = amp * climit_inv; // amp must fit for limiting : amp(new) = limit/maxval; = amp(old)*limitOUT/tresh;
+	  else *out++ = amp * climit_inv; /* amp must fit for limiting : amp(new) = limit/maxval; = amp(old)*limitOUT/tresh; */
 	else *out++ = 1.;
 
 	*in++ = 0.;
@@ -561,7 +622,7 @@ static t_int *limiter_perform(t_int *w)
     break;
   }
 
-  // now return the goodies
+  /* now return the goodies */
   x->amplification	= amp;
   x->samples_left		= samplesleft;
 
@@ -569,27 +630,10 @@ static t_int *limiter_perform(t_int *w)
 }
 
 
-#if 0
-static t_int *route_through(t_int *w)
-{
-  t_float *in  = (t_float *)w[1];
-  t_float *out = (t_float *)w[2];
-  int n = (int)w[3];
-
-  while(n--)
-    {
-      *out++ = *in;
-      *in++ = 0;
-    }
-
-  return (w+4);
-}
-#endif
-
-void limiter_dsp(t_limiter *x, t_signal **sp)
+static void limiter_dsp(t_limiter *x, t_signal **sp)
 {
   int i = 0;
-  t_float* sig_buf = (t_float *)getbytes(sizeof(t_float) * sp[0]->s_n);
+  t_sample* sig_buf = (t_sample *)getbytes(sizeof(*sig_buf) * sp[0]->s_n);
 
   x->s_n = sp[0]->s_n;
 
@@ -609,12 +653,11 @@ void limiter_dsp(t_limiter *x, t_signal **sp)
 
 
 /* ------------------------------------------------------------------------------------ */
-// finally do the creation - things
+/* finally do the creation - things */
 
-void *limiter_new(t_symbol *s, int argc, t_atom *argv)
+static void *limiter_new(t_symbol *s, int argc, t_atom *argv)
 {
   t_limiter *x = (t_limiter *)pd_new(limiter_class);
-
   int i = 0;
 
   if (argc) set_bufsize(x, atom_getfloat(argv));
@@ -640,7 +683,7 @@ void *limiter_new(t_symbol *s, int argc, t_atom *argv)
   while (i < x->number_of_inlets)
     {
       int n;
-      t_float* buf = (float *)getbytes(sizeof(float) * x->buf_size);
+      t_sample* buf = (t_sample *)getbytes(sizeof(*buf) * x->buf_size);
       x->in[i].ringbuf		= buf;
       x->in[i].buf_position	= 0;
       for (n = 0; n < x->buf_size; n++) x->in[i].ringbuf[n] = 0.;
@@ -663,7 +706,7 @@ void *limiter_new(t_symbol *s, int argc, t_atom *argv)
   return (x);
 }
 
-void limiter_free(t_limiter *x)
+static void limiter_free(t_limiter *x)
 {
   int i=0;
 
@@ -671,7 +714,7 @@ void limiter_free(t_limiter *x)
   freebytes(x->val2, sizeof(t_limctl));
   freebytes(x->cmp , sizeof(t_cmpctl));
 
-  while (i < x->number_of_inlets)	freebytes(x->in[i++].ringbuf, x->buf_size * sizeof(t_float));
+  while (i < x->number_of_inlets)	freebytes(x->in[i++].ringbuf, x->buf_size * sizeof(t_sample));
 
   freebytes(x->in, x->number_of_inlets * sizeof(t_inbuf));
 }
@@ -685,15 +728,16 @@ void limiter_free(t_limiter *x)
 
 void limiter_tilde_setup(void)
 {
+  init_sinc();
+
   limiter_class = class_new(gensym("limiter~"), (t_newmethod)limiter_new, (t_method)limiter_free,
 			    sizeof(t_limiter), 0, A_GIMME, 0);
 
   class_addmethod(limiter_class, nullfn,					gensym("signal"), 0);
   class_addmethod(limiter_class, (t_method)limiter_dsp,	gensym("dsp"), 0);
 
-  class_addmethod(limiter_class, (t_method)helper,	gensym("help"), 0);
+  class_addmethod(limiter_class, (t_method)limiter_tilde_helper,	gensym("help"), 0);
   class_addmethod(limiter_class, (t_method)status,	gensym("print"), 0);
-  class_sethelpsymbol(limiter_class, gensym("zexy/limiter~"));
 
   class_addmethod(limiter_class, (t_method)set_mode,	gensym("mode"), A_FLOAT, 0);
   class_addmethod(limiter_class, (t_method)set_LIMIT,	gensym("LIMIT"),  0);
@@ -716,8 +760,3 @@ void limiter_tilde_setup(void)
 
   zexy_register("limiter~");
 }
-
-void z_limiter__setup(void)
-{
-  limiter_tilde_setup();
-}
